@@ -1,3 +1,4 @@
+#![allow(clippy::type_complexity)]
 use regex::Regex;
 use std::io::prelude::*;
 use std::sync::atomic::{AtomicUsize, Ordering};
@@ -20,7 +21,7 @@ use std::{
 };
 
 use chrono::prelude::*;
-use chrono::DateTime;
+use chrono::{DateTime, TimeZone, Utc};
 use crypto_msg_parser::{extract_symbol, parse_l2, parse_trade, MarketType, MessageType};
 use dashmap::{DashMap, DashSet};
 use flate2::write::GzEncoder;
@@ -50,20 +51,17 @@ pub struct Message {
 }
 
 fn get_day(timestamp_millis: i64) -> String {
-    let naive = NaiveDateTime::from_timestamp(timestamp_millis / 1000, 0);
-    let datetime: DateTime<Utc> = DateTime::from_utc(naive, Utc);
-    datetime.format("%Y-%m-%d").to_string()
+    let dt = Utc.timestamp(timestamp_millis / 1000, 0);
+    dt.format("%Y-%m-%d").to_string()
 }
 
 fn get_hour(timestamp_millis: i64) -> String {
-    let naive = NaiveDateTime::from_timestamp(timestamp_millis / 1000, 0);
-    let datetime: DateTime<Utc> = DateTime::from_utc(naive, Utc);
-    datetime.format("%Y-%m-%d-%H").to_string()
+    let dt = Utc.timestamp(timestamp_millis / 1000, 0);
+    dt.format("%Y-%m-%d-%H").to_string()
 }
 
 // Output to a raw file and a parsed file.
 #[derive(Clone)]
-#[allow(clippy::type_complexity)]
 struct Output(Arc<Mutex<Box<dyn std::io::Write + Send>>>);
 
 // Copied from crypto-markets/tests/bitmex.rs
@@ -116,20 +114,19 @@ fn is_blocked_market(market_type: MarketType) -> bool {
 ///
 /// - input_file A `.json.gz` file downloaded from AWS S3
 /// - day `yyyy-MM-dd` string, all messages beyond [day-5min, day+5min] will be dropped
-/// - output_dir_raw Where raw messages will be written to
-/// - output_dir_parsed Where parsed messages will be written to
-/// - split_files A HashMap that tracks opened files, key is `msg.symbol`, value is file of
+/// - output_dir Where raw messages will be written to
+/// - splitted_files A HashMap that tracks opened files, key is `msg.symbol`, value is file of
 ///  `output_dir/exchange.market_type.msg_type.symbol.day.json.gz`. Each `exchange, msg_type, market_type`
-///  has one `split_files` HashMap
-/// - visited A HashSet for deduplication, each `exchange, msg_type, market_type` has one `visited` Hashset
-fn split_file<P>(
+///  has one `splitted_files` HashMap
+/// - visited A HashSet for deduplication, each `exchange, msg_type, market_type` has one
+///  `visited` Hashset
+fn split_file_raw<P>(
     input_file: P,
     day: String,
-    output_dir_raw: P,
-    output_dir_parsed: P,
-    split_files: Arc<DashMap<String, Output>>,
+    output_dir: P,
+    splitted_files: Arc<DashMap<String, Output>>,
     visited: Arc<DashSet<u64>>,
-) -> (i64, i64, i64)
+) -> (i64, i64, i64, i64, i64)
 where
     P: AsRef<Path>,
 {
@@ -143,8 +140,10 @@ where
         .unwrap_or_else(|_| panic!("{:?} does not exist", input_file.as_ref().display()));
     let buf_reader = std::io::BufReader::new(GzDecoder::new(f_in));
     let mut total_lines = 0;
-    let mut new_lines = 0;
+    let mut unique_lines = 0;
+    let mut duplicated_lines = 0;
     let mut error_lines = 0;
+    let mut expired_lines = 0;
     let re = Regex::new(r"[():.\\/]+").unwrap();
     for line in buf_reader.lines() {
         if let Ok(line) = line {
@@ -155,116 +154,188 @@ where
                     assert_eq!(msg.market_type, market_type);
                 }
                 assert_eq!(msg.msg_type, msg_type);
-                let is_new = {
-                    let hashcode = {
-                        let mut hasher = DefaultHasher::new();
-                        msg.json.hash(&mut hasher);
-                        hasher.finish()
-                    };
-                    visited.insert(hashcode)
+                let hashcode = {
+                    let mut hasher = DefaultHasher::new();
+                    msg.json.hash(&mut hasher);
+                    hasher.finish()
                 };
-                if is_new {
-                    new_lines += 1;
-                    if let Some(symbol) = extract_symbol(exchange, market_type, &msg.json) {
-                        let real_market_type =
-                            get_real_market_type(exchange, msg.market_type, &symbol);
+                if let Some(symbol) = extract_symbol(exchange, market_type, &msg.json) {
+                    let real_market_type = get_real_market_type(exchange, msg.market_type, &symbol);
+
+                    if day == get_day(msg.received_at as i64) {
                         // raw
-                        if day == get_day(msg.received_at as i64) {
-                            let hour = get_hour(msg.received_at as i64);
-                            let key = format!("raw.{}.{}.{}", real_market_type, symbol, hour);
-                            if !split_files.contains_key(&key) {
-                                let buf_writer_raw = {
-                                    let output_file_name = format!(
-                                        "{}.{}.{}.{}.{}.json.gz",
-                                        exchange,
-                                        real_market_type,
-                                        msg_type_str,
-                                        re.replace_all(&symbol, "_"),
-                                        hour
-                                    );
-                                    let output_dir = Path::new(output_dir_raw.as_ref())
-                                        .join(real_market_type.to_string());
-                                    std::fs::create_dir_all(output_dir.as_path()).unwrap();
-                                    let f_out = std::fs::OpenOptions::new()
-                                        .create(true)
-                                        .write(true)
-                                        .truncate(true)
-                                        .open(
-                                            Path::new(output_dir.as_path()).join(output_file_name),
-                                        )
-                                        .unwrap();
-                                    std::io::BufWriter::new(GzEncoder::new(
-                                        f_out,
-                                        Compression::default(),
-                                    ))
-                                };
-                                split_files.insert(
-                                    key.clone(),
-                                    Output(Arc::new(Mutex::new(Box::new(buf_writer_raw)))),
-                                );
-                            }
+                        if visited.insert(hashcode) {
+                            unique_lines += 1;
+                            let output_file_name = {
+                                let hour = get_hour(msg.received_at as i64);
+                                format!(
+                                    "{}.{}.{}.{}.{}.json.gz",
+                                    exchange,
+                                    real_market_type,
+                                    msg_type_str,
+                                    re.replace_all(&symbol, "_"),
+                                    hour
+                                )
+                            };
                             let output = {
-                                let entry = split_files.get(&key).unwrap();
+                                let output_dir_clone = output_dir.as_ref().to_path_buf();
+                                // `.entry().or_insert_with()` is atomic, see https://github.com/xacrimon/dashmap/issues/78
+                                let entry = splitted_files
+                                    .entry(output_file_name.clone())
+                                    .or_insert_with(move || {
+                                        let buf_writer_raw = {
+                                            let output_dir =
+                                                output_dir_clone.join(real_market_type.to_string());
+                                            std::fs::create_dir_all(output_dir.as_path()).unwrap();
+                                            let f_out = std::fs::OpenOptions::new()
+                                                .create(true)
+                                                .write(true)
+                                                .truncate(true)
+                                                .open(
+                                                    Path::new(output_dir.as_path())
+                                                        .join(output_file_name),
+                                                )
+                                                .unwrap();
+                                            std::io::BufWriter::new(GzEncoder::new(
+                                                f_out,
+                                                Compression::default(),
+                                            ))
+                                        };
+                                        Output(Arc::new(Mutex::new(Box::new(buf_writer_raw))))
+                                    });
                                 entry.value().clone()
                             };
+                            let mut writer = output.0.lock().unwrap();
                             if msg.market_type == MarketType::Unknown {
                                 msg.market_type = real_market_type;
-                                writeln!(
-                                    output.0.lock().unwrap(),
-                                    "{}",
-                                    serde_json::to_string(&msg).unwrap()
-                                )
-                                .unwrap();
+                                writeln!(writer, "{}", serde_json::to_string(&msg).unwrap())
+                                    .unwrap();
                             } else {
-                                writeln!(output.0.lock().unwrap(), "{}", line).unwrap();
+                                assert_eq!(real_market_type, msg.market_type);
+                                writeln!(writer, "{}", line).unwrap();
                             }
+                        } else {
+                            duplicated_lines += 1;
                         }
+                    } else {
+                        expired_lines += 1;
+                    }
+                } else {
+                    warn!("No symbol: {}", line);
+                    error_lines += 1;
+                }
+            } else {
+                warn!("{}", line);
+                error_lines += 1;
+            }
+        } else {
+            error!("malformed file {}", input_file.as_ref().display());
+            error_lines += 1;
+        }
+    }
+    (
+        total_lines,
+        unique_lines,
+        duplicated_lines,
+        error_lines,
+        expired_lines,
+    )
+}
+
+fn split_file_parsed<P>(
+    input_file: P,
+    day: String,
+    output_dir: P,
+    splitted_files: Arc<DashMap<String, Output>>,
+    visited: Arc<DashSet<u64>>,
+) -> (i64, i64, i64, i64, i64)
+where
+    P: AsRef<Path>,
+{
+    let file_name = input_file.as_ref().file_name().unwrap();
+    let v: Vec<&str> = file_name.to_str().unwrap().split('.').collect();
+    let exchange = v[0];
+    let market_type = MarketType::from_str(v[1]).unwrap();
+    let msg_type_str = v[2];
+    let msg_type = MessageType::from_str(msg_type_str).unwrap();
+    let f_in = std::fs::File::open(&input_file)
+        .unwrap_or_else(|_| panic!("{:?} does not exist", input_file.as_ref().display()));
+    let buf_reader = std::io::BufReader::new(GzDecoder::new(f_in));
+    let mut total_lines = 0;
+    let mut unique_lines = 0;
+    let mut duplicated_lines = 0;
+    let mut error_lines = 0;
+    let mut expired_lines = 0;
+    let re = Regex::new(r"[():.\\/]+").unwrap();
+    for line in buf_reader.lines() {
+        if let Ok(line) = line {
+            total_lines += 1;
+            if let Ok(msg) = serde_json::from_str::<Message>(&line) {
+                assert_eq!(msg.exchange, exchange);
+                if market_type != MarketType::Unknown {
+                    assert_eq!(msg.market_type, market_type);
+                }
+                assert_eq!(msg.msg_type, msg_type);
+                let hashcode = {
+                    let mut hasher = DefaultHasher::new();
+                    msg.json.hash(&mut hasher);
+                    hasher.finish()
+                };
+                if let Some(symbol) = extract_symbol(exchange, market_type, &msg.json) {
+                    let real_market_type = get_real_market_type(exchange, msg.market_type, &symbol);
+
+                    if visited.insert(hashcode) {
+                        unique_lines += 1;
                         // parsed
                         let write_parsed =
                             |market_type: MarketType, json: String, timestamp: i64| {
-                                let hour = get_hour(timestamp);
-                                let key = format!("parsed.{}.{}.{}", market_type, symbol, hour);
-                                if !split_files.contains_key(&key) {
-                                    let buf_writer_parsed = {
-                                        let pair =
-                                            crypto_pair::normalize_pair(&symbol, exchange).unwrap();
-                                        let pair = re.replace_all(&pair, "_");
-                                        let output_file_name = format!(
-                                            "{}.{}.{}.{}.{}.{}.json.gz",
-                                            exchange,
-                                            market_type,
-                                            msg_type_str,
-                                            re.replace_all(&pair, "_"),
-                                            re.replace_all(&symbol, "_"),
-                                            hour
-                                        );
-                                        let output_dir = Path::new(output_dir_parsed.as_ref())
-                                            .join(market_type.to_string());
-                                        std::fs::create_dir_all(output_dir.as_path()).unwrap();
-                                        let f_out = std::fs::OpenOptions::new()
-                                            .create(true)
-                                            .write(true)
-                                            .truncate(true)
-                                            .open(
-                                                Path::new(output_dir.as_path())
-                                                    .join(output_file_name),
-                                            )
-                                            .unwrap();
-                                        std::io::BufWriter::new(GzEncoder::new(
-                                            f_out,
-                                            Compression::default(),
-                                        ))
-                                    };
-                                    split_files.insert(
-                                        key.clone(),
-                                        Output(Arc::new(Mutex::new(Box::new(buf_writer_parsed)))),
-                                    );
-                                }
+                                let output_file_name = {
+                                    let hour = get_hour(timestamp);
+                                    let pair =
+                                        crypto_pair::normalize_pair(&symbol, exchange).unwrap();
+                                    format!(
+                                        "{}.{}.{}.{}.{}.{}.json.gz",
+                                        exchange,
+                                        market_type,
+                                        msg_type_str,
+                                        re.replace_all(&pair, "_"),
+                                        re.replace_all(&symbol, "_"),
+                                        hour
+                                    )
+                                };
                                 let output = {
-                                    let entry = split_files.get(&key).unwrap();
+                                    let output_dir_clone = output_dir.as_ref().to_path_buf();
+                                    // `.entry().or_insert_with()` is atomic, see https://github.com/xacrimon/dashmap/issues/78
+                                    let entry = splitted_files
+                                        .entry(output_file_name.clone())
+                                        .or_insert_with(move || {
+                                            let buf_writer_parsed = {
+                                                let output_dir =
+                                                    output_dir_clone.join(market_type.to_string());
+                                                std::fs::create_dir_all(output_dir.as_path())
+                                                    .unwrap();
+                                                let f_out = std::fs::OpenOptions::new()
+                                                    .create(true)
+                                                    .write(true)
+                                                    .truncate(true)
+                                                    .open(
+                                                        Path::new(output_dir.as_path())
+                                                            .join(output_file_name),
+                                                    )
+                                                    .unwrap();
+                                                std::io::BufWriter::new(GzEncoder::new(
+                                                    f_out,
+                                                    Compression::default(),
+                                                ))
+                                            };
+                                            Output(Arc::new(Mutex::new(Box::new(
+                                                buf_writer_parsed,
+                                            ))))
+                                        });
                                     entry.value().clone()
                                 };
-                                writeln!(output.0.lock().unwrap(), "{}", json).unwrap();
+                                let mut writer = output.0.lock().unwrap();
+                                writeln!(writer, "{}", json).unwrap();
                             };
 
                         match msg.msg_type {
@@ -278,12 +349,15 @@ where
                                         Some(msg.received_at as i64),
                                     ) {
                                         for message in messages {
+                                            assert_eq!(real_market_type, message.market_type);
                                             if get_day(message.timestamp) == day {
                                                 write_parsed(
                                                     message.market_type,
                                                     serde_json::to_string(&message).unwrap(),
                                                     message.timestamp,
                                                 );
+                                            } else {
+                                                expired_lines += 1;
                                             }
                                         }
                                     }
@@ -294,12 +368,15 @@ where
                                     parse_trade(&msg.exchange, msg.market_type, &msg.json)
                                 {
                                     for message in messages {
+                                        assert_eq!(real_market_type, message.market_type);
                                         if get_day(message.timestamp) == day {
                                             write_parsed(
                                                 message.market_type,
                                                 serde_json::to_string(&message).unwrap(),
                                                 message.timestamp,
                                             );
+                                        } else {
+                                            expired_lines += 1;
                                         }
                                     }
                                 }
@@ -307,9 +384,11 @@ where
                             _ => panic!("Unknown msg_type {}", msg.msg_type),
                         };
                     } else {
-                        warn!("{}", line);
-                        error_lines += 1;
+                        duplicated_lines += 1;
                     }
+                } else {
+                    warn!("No symbol: {}", line);
+                    error_lines += 1;
                 }
             } else {
                 warn!("{}", line);
@@ -320,7 +399,13 @@ where
             error_lines += 1;
         }
     }
-    (total_lines, new_lines, error_lines)
+    (
+        total_lines,
+        unique_lines,
+        duplicated_lines,
+        error_lines,
+        expired_lines,
+    )
 }
 
 fn sort_file<P>(input_file: P, writer: &mut dyn std::io::Write) -> (i64, i64)
@@ -445,7 +530,7 @@ where
                 // Wait for the semaphore which allows only two pixz processes
                 let mut rng = rand::thread_rng();
                 while semaphore.load(Ordering::SeqCst) == 0 {
-                    let millis = rng.gen_range(1000_u64..5000_u64);
+                    let millis = rng.gen_range(3000_u64..10000_u64);
                     debug!("Waiting for semaphore");
                     std::thread::sleep(Duration::from_millis(millis));
                 }
@@ -566,15 +651,25 @@ fn process_files_of_day(
             "Started split {} {} {} {}",
             exchange, market_type, msg_type, day
         );
-        #[allow(clippy::type_complexity)]
-        let (tx, rx): (Sender<(i64, i64, i64)>, Receiver<(i64, i64, i64)>) = mpsc::channel();
+        let (tx_raw, rx_raw): (
+            Sender<(i64, i64, i64, i64, i64)>,
+            Receiver<(i64, i64, i64, i64, i64)>,
+        ) = mpsc::channel();
+        let (tx_parsed, rx_parsed): (
+            Sender<(i64, i64, i64, i64, i64)>,
+            Receiver<(i64, i64, i64, i64, i64)>,
+        ) = mpsc::channel();
         let start_timstamp = Instant::now();
         // Larger files get processed first
         paths.sort_by_cached_key(|path| Reverse(std::fs::metadata(path).unwrap().len()));
-        let visited: Arc<DashSet<u64>> = Arc::new(DashSet::new());
-        let split_files: Arc<DashMap<String, Output>> = Arc::new(DashMap::new());
-        for path in paths {
-            let file_name = path.as_path().file_name().unwrap();
+
+        let written_to_raw: Arc<DashSet<u64>> = Arc::new(DashSet::new());
+        let written_to_parsed: Arc<DashSet<u64>> = Arc::new(DashSet::new());
+        let splitted_files_raw: Arc<DashMap<String, Output>> = Arc::new(DashMap::new());
+        let splitted_files_parsed: Arc<DashMap<String, Output>> = Arc::new(DashMap::new());
+
+        for input_file in paths {
+            let file_name = input_file.as_path().file_name().unwrap();
             let v: Vec<&str> = file_name.to_str().unwrap().split('.').collect();
             assert_eq!(exchange, v[0]);
             if market_type != MarketType::Unknown {
@@ -582,59 +677,99 @@ fn process_files_of_day(
             }
             let msg_type_str = v[2];
             assert_eq!(msg_type, MessageType::from_str(msg_type_str).unwrap());
-            let output_dir_raw = Path::new(output_dir_raw).join(msg_type_str).join(exchange);
-            std::fs::create_dir_all(output_dir_raw.as_path()).unwrap();
-            let output_dir_parsed = Path::new(output_dir_parsed)
+
+            let input_file_clone = input_file.clone();
+            let day_clone = day.to_string();
+            let exchange_output_dir_raw =
+                Path::new(output_dir_raw).join(msg_type_str).join(exchange);
+            let splitted_files_raw_clone = splitted_files_raw.clone();
+            let written_to_raw_clone = written_to_raw.clone();
+            let tx_raw_clone = tx_raw.clone();
+            thread_pool.execute(move || {
+                let t = split_file_raw(
+                    input_file_clone,
+                    day_clone,
+                    exchange_output_dir_raw,
+                    splitted_files_raw_clone,
+                    written_to_raw_clone,
+                );
+                tx_raw_clone.send(t).unwrap();
+            });
+
+            let input_file_clone = input_file.clone();
+            let day_clone = day.to_string();
+            let splitted_files_parsed_clone = splitted_files_parsed.clone();
+            let exchange_output_dir_parsed = Path::new(output_dir_parsed)
                 .join(msg_type_str)
                 .join(exchange);
-            std::fs::create_dir_all(output_dir_parsed.as_path()).unwrap();
-
-            let visited_clone = visited.clone();
-            let split_files_clone = split_files.clone();
-            let day_clone = day.to_string();
-            let thread_tx = tx.clone();
+            let written_to_parsed_clone = written_to_parsed.clone();
+            let tx_parsed_clone = tx_parsed.clone();
             thread_pool.execute(move || {
-                let t = split_file(
-                    path.as_path(),
+                let t = split_file_parsed(
+                    input_file_clone,
                     day_clone,
-                    output_dir_raw.as_path(),
-                    output_dir_parsed.as_path(),
-                    split_files_clone,
-                    visited_clone,
+                    exchange_output_dir_parsed,
+                    splitted_files_parsed_clone,
+                    written_to_parsed_clone,
                 );
-                thread_tx.send(t).unwrap();
+                tx_parsed_clone.send(t).unwrap();
             });
         }
         thread_pool.join();
-        drop(tx); // drop the sender
-        let mut total_lines = 0;
-        let mut new_lines = 0;
-        let mut error_lines = 0;
-        for t in rx {
-            total_lines += t.0;
-            new_lines += t.1;
-            error_lines += t.2;
-        }
-        for entry in split_files.iter() {
-            let output = entry.value();
-            output.0.lock().unwrap().flush().unwrap();
-        }
-        let error_ratio = (error_lines as f64) / (total_lines as f64);
-        if error_ratio > 0.01 {
-            // error ratio > 1%
-            error!(
-                "Failed to split {} {} {} {}, because error ratio {}/{}={}% is higher than 1% !",
-                exchange,
-                market_type,
-                msg_type,
-                day,
-                error_lines,
-                total_lines,
-                error_ratio * 100.0
-            );
+        let finishing = move |tx: Sender<(i64, i64, i64, i64, i64)>,
+                              rx: Receiver<(i64, i64, i64, i64, i64)>,
+                              splitted_files: Arc<DashMap<String, Output>>,
+                              is_parsed: bool|
+              -> bool {
+            drop(tx); // drop the sender to unblock receiver
+            let mut total_lines = 0;
+            let mut unique_lines = 0;
+            let mut duplicated_lines = 0;
+            let mut error_lines = 0;
+            let mut expired_lines = 0;
+            for t in rx {
+                total_lines += t.0;
+                unique_lines += t.1;
+                duplicated_lines += t.2;
+                error_lines += t.3;
+                expired_lines += t.4;
+            }
+            if is_parsed {
+                assert_eq!(total_lines, unique_lines + duplicated_lines + error_lines);
+            } else {
+                assert_eq!(
+                    total_lines,
+                    unique_lines + duplicated_lines + error_lines + expired_lines
+                );
+            }
+
+            for entry in splitted_files.iter() {
+                let output = entry.value();
+                output.0.lock().unwrap().flush().unwrap();
+            }
+            let error_ratio = (error_lines as f64) / (total_lines as f64);
+            if error_ratio > 0.01 {
+                // error ratio > 1%
+                error!(
+                    "Failed to split {} {} {} {}, because error ratio {}/{}={}% is higher than 1% !",
+                    exchange,
+                    market_type,
+                    msg_type,
+                    day,
+                    error_lines,
+                    total_lines,
+                    error_ratio * 100.0
+                );
+                false
+            } else {
+                info!("Finished split {} {} {} {}, total {} lines, {} unique lines, {} duplicated lines, {} expired lines,  {} malformed lines, time elapsed {} seconds", exchange, market_type, msg_type, day, total_lines, unique_lines, duplicated_lines, expired_lines, error_lines, start_timstamp.elapsed().as_secs());
+                true
+            }
+        };
+        if !finishing(tx_raw, rx_raw, splitted_files_raw, false)
+            || !finishing(tx_parsed, rx_parsed, splitted_files_parsed, true)
+        {
             return false;
-        } else {
-            info!("Finished split {} {} {} {}, total {} lines, {} unique lines, {} malformed lines, time elapsed {} seconds", exchange, market_type, msg_type, day, total_lines, new_lines, error_lines, start_timstamp.elapsed().as_secs());
         }
     }
 
@@ -695,7 +830,6 @@ fn process_files_of_day(
             "Started sort {} {} {} {}",
             exchange, market_type, msg_type, day
         );
-        #[allow(clippy::type_complexity)]
         let (tx, rx): (Sender<(i64, i64)>, Receiver<(i64, i64)>) = mpsc::channel();
         let start_timstamp = Instant::now();
         let percentile_90 = ((paths_by_day.len() as f64) * 0.9) as usize;
